# Copyright (c) 2013 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import array
import binascii
import os

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_log import versionutils

from os_brick.encryptors import base
from os_brick import exception

LOG = logging.getLogger(__name__)


class CryptsetupEncryptor(base.VolumeEncryptor):
    """A VolumeEncryptor based on dm-crypt.

    This VolumeEncryptor uses dm-crypt to encrypt the specified volume.
    """

    def __init__(self, root_helper,
                 connection_info,
                 keymgr,
                 execute=None,
                 *args, **kwargs):
        super(CryptsetupEncryptor, self).__init__(
            root_helper=root_helper,
            connection_info=connection_info,
            keymgr=keymgr,
            execute=execute,
            *args, **kwargs)

        # Fail if no device_path was set when connecting the volume, e.g. in
        # the case of libvirt network volume drivers.
        data = connection_info['data']
        if not data.get('device_path'):
            volume_id = data.get('volume_id') or connection_info.get('serial')
            raise exception.VolumeEncryptionNotSupported(
                volume_id=volume_id,
                volume_type=connection_info['driver_volume_type'])

        # the device's path as given to libvirt -- e.g., /dev/disk/by-path/...
        self.symlink_path = connection_info['data']['device_path']

        # a unique name for the volume -- e.g., the iSCSI participant name
        self.dev_name = 'crypt-%s' % os.path.basename(self.symlink_path)

        # NOTE(lixiaoy1): This is to import fix for 1439869 from Nova.
        # NOTE(tsekiyama): In older version of nova, dev_name was the same
        # as the symlink name. Now it has 'crypt-' prefix to avoid conflict
        # with multipath device symlink. To enable rolling update, we use the
        # old name when the encrypted volume already exists.
        old_dev_name = os.path.basename(self.symlink_path)
        wwn = data.get('multipath_id')
        if self._is_crypt_device_available(old_dev_name):
            self.dev_name = old_dev_name
            LOG.debug("Using old encrypted volume name: %s", self.dev_name)
        elif wwn and wwn != old_dev_name:
            # FibreChannel device could be named '/dev/mapper/<WWN>'.
            if self._is_crypt_device_available(wwn):
                self.dev_name = wwn
                LOG.debug("Using encrypted volume name from wwn: %s",
                          self.dev_name)

        # the device's actual path on the compute host -- e.g., /dev/sd_
        self.dev_path = os.path.realpath(self.symlink_path)

    def _is_crypt_device_available(self, dev_name):
        if not os.path.exists('/dev/mapper/%s' % dev_name):
            return False

        try:
            self._execute('cryptsetup', 'status', dev_name, run_as_root=True)
        except processutils.ProcessExecutionError as e:
            # If /dev/mapper/<dev_name> is a non-crypt block device (such as a
            # normal disk or multipath device), exit_code will be 1. In the
            # case, we will omit the warning message.
            if e.exit_code != 1:
                LOG.warning('cryptsetup status %(dev_name)s exited '
                            'abnormally (status %(exit_code)s): %(err)s',
                            {"dev_name": dev_name, "exit_code": e.exit_code,
                             "err": e.stderr})
            return False
        return True

    def _get_passphrase(self, key):
        """Convert raw key to string."""
        return binascii.hexlify(key).decode('utf-8')

    def _open_volume(self, passphrase, **kwargs):
        """Open the LUKS partition on the volume using passphrase.

        :param passphrase: the passphrase used to access the volume
        """
        LOG.debug("opening encrypted volume %s", self.dev_path)

        # NOTE(joel-coffman): cryptsetup will strip trailing newlines from
        # input specified on stdin unless --key-file=- is specified.
        cmd = ["cryptsetup", "create", "--key-file=-"]

        cipher = kwargs.get("cipher", None)
        if cipher is not None:
            cmd.extend(["--cipher", cipher])

        key_size = kwargs.get("key_size", None)
        if key_size is not None:
            cmd.extend(["--key-size", key_size])

        cmd.extend([self.dev_name, self.dev_path])

        self._execute(*cmd, process_input=passphrase,
                      check_exit_code=True, run_as_root=True,
                      root_helper=self._root_helper)

    def _get_mangled_passphrase(self, key):
        """Convert the raw key into a list of unsigned int's and then a string

        """
        # NOTE(lyarwood): This replicates the methods used prior to Newton to
        # first encode the passphrase as a list of unsigned int's before
        # decoding back into a string. This method strips any leading 0's
        # of the resulting hex digit pairs, resulting in a different
        # passphrase being returned.
        encoded_key = array.array('B', key).tolist()
        return ''.join(hex(x).replace('0x', '') for x in encoded_key)

    def attach_volume(self, context, **kwargs):
        """Shadow the device and pass an unencrypted version to the instance.

        Transparent disk encryption is achieved by mounting the volume via
        dm-crypt and passing the resulting device to the instance. The
        instance is unaware of the underlying encryption due to modifying the
        original symbolic link to refer to the device mounted by dm-crypt.
        """
        # TODO(lyarwood): Remove this encryptor and refactor the LUKS based
        # encryptors in the U release.
        versionutils.report_deprecated_feature(
            LOG,
            "The plain CryptsetupEncryptor is deprecated and will be removed "
            "in a future release. Existing users are encouraged to retype "
            "any existing volumes using this encryptor to the 'luks' "
            "LuksEncryptor or 'luks2' Luks2Encryptor encryptors as soon as "
            "possible.")
        key = self._get_key(context).get_encoded()
        passphrase = self._get_passphrase(key)

        try:
            self._open_volume(passphrase, **kwargs)
        except processutils.ProcessExecutionError as e:
            if e.exit_code == 2:
                # NOTE(lyarwood): Workaround bug#1633518 by attempting to use
                # a mangled passphrase to open the device..
                LOG.info("Unable to open %s with the current passphrase, "
                         "attempting to use a mangled passphrase to open "
                         "the volume.", self.dev_path)
                self._open_volume(self._get_mangled_passphrase(key), **kwargs)

        # modify the original symbolic link to refer to the decrypted device
        self._execute('ln', '--symbolic', '--force',
                      '/dev/mapper/%s' % self.dev_name, self.symlink_path,
                      root_helper=self._root_helper,
                      run_as_root=True, check_exit_code=True)

    def _close_volume(self, **kwargs):
        """Closes the device (effectively removes the dm-crypt mapping)."""
        LOG.debug("closing encrypted volume %s", self.dev_path)
        # NOTE(mdbooth): remove will return 4 (wrong device specified) if
        # the device doesn't exist. We assume here that the caller hasn't
        # specified the wrong device, and that it doesn't exist because it
        # isn't open. We don't fail in this case in order to make this
        # operation idempotent.
        self._execute('cryptsetup', 'remove', self.dev_name,
                      run_as_root=True, check_exit_code=[0, 4],
                      root_helper=self._root_helper)

    def detach_volume(self, **kwargs):
        """Removes the dm-crypt mapping for the device."""
        self._close_volume(**kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import abc

from oslo_log import log as logging
from oslo_utils import importutils

from os_brick import exception
from os_brick.i18n import _


LOG = logging.getLogger(__name__)

CACHE_ENGINE_TO_CACHE_CLASS_MAP = {
    "opencas": 'os_brick.caches.opencas.OpenCASEngine',
}


class CacheEngineBase(object, metaclass=abc.ABCMeta):
    def __init__(self, **kwargs):
        self._root_helper = kwargs.get('root_helper')

    @abc.abstractmethod
    def is_engine_ready(self, **kwargs):
        return

    @abc.abstractmethod
    def attach_volume(self, **kwargs):
        return

    @abc.abstractmethod
    def detach_volume(self, **kwargs):
        return


class CacheManager():
    """Cache manager for volumes.

    This CacheManager uses cache engines to do volume cache.
    """
    def __init__(self, root_helper, connection_info,
                 *args, **kwargs):

        data = connection_info['data']
        if not data.get('device_path'):
            volume_id = data.get('volume_id') or connection_info.get('serial')
            raise exception.VolumeLocalCacheNotSupported(
                volume_id=volume_id,
                volume_type=connection_info.get('driver_volume_type'))

        self.ori_device_path = data.get('device_path')
        if not data.get('cacheable'):
            self.cacheable = False
            return

        self.cacheable = True
        self.root_helper = root_helper
        self.engine_name = kwargs.get('cache_name')
        self.args = args
        self.kwargs = kwargs
        self.kwargs["root_helper"] = root_helper
        self.kwargs["dev_path"] = data.get('device_path')
        self.engine = self._get_engine(self.engine_name, **self.kwargs)

    def _get_engine(self, engine_name, **kwargs):
        eng_cls_path = CACHE_ENGINE_TO_CACHE_CLASS_MAP.get(engine_name)
        if eng_cls_path:
            engine_cls = importutils.import_class(eng_cls_path)
            eng = engine_cls(**kwargs)
            if eng.is_engine_ready():
                return eng

        raise exception.Invalid(_("No valid cache engine"))

    def attach_volume(self):
        """setup the cache when attaching volume."""
        if not self.cacheable:
            return self.ori_device_path

        LOG.debug("volume before cached: %s", self.kwargs.get('dev_path'))
        emulated_disk = self.engine.attach_volume(**self.kwargs)
        LOG.debug("volume after cached: %s", emulated_disk)
        return emulated_disk

    def detach_volume(self):
        """Release the cache on detaching volume."""
        if not self.cacheable:
            return self.ori_device_path

        LOG.debug("volume before detach: %s", self.kwargs.get('dev_path'))
        ori_disk = self.engine.detach_volume(**self.kwargs)
        LOG.debug("volume after detach: %s", ori_disk)
        return ori_disk



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Windows remote filesystem client utilities."""

import os
import re

from os_win import utilsfactory
from oslo_log import log as logging

from os_brick import exception
from os_brick.i18n import _
from os_brick.remotefs import remotefs

LOG = logging.getLogger(__name__)


class WindowsRemoteFsClient(remotefs.RemoteFsClient):
    _username_regex = re.compile(r'user(?:name)?=([^, ]+)')
    _password_regex = re.compile(r'pass(?:word)?=([^, ]+)')
    _loopback_share_map = {}

    def __init__(self, mount_type, root_helper=None,
                 execute=None, *args, **kwargs):
        mount_type_to_option_prefix = {
            'cifs': 'smbfs',
            'smbfs': 'smbfs',
        }

        self._local_path_for_loopback = kwargs.get('local_path_for_loopback',
                                                   True)

        if mount_type not in mount_type_to_option_prefix:
            raise exception.ProtocolNotSupported(protocol=mount_type)

        self._mount_type = mount_type
        option_prefix = mount_type_to_option_prefix[mount_type]

        self._mount_base = kwargs.get(option_prefix + '_mount_point_base')
        self._mount_options = kwargs.get(option_prefix + '_mount_options')

        self._smbutils = utilsfactory.get_smbutils()
        self._pathutils = utilsfactory.get_pathutils()

    def get_local_share_path(self, share, expect_existing=True):
        share = self._get_share_norm_path(share)
        share_name = self.get_share_name(share)
        share_subdir = self.get_share_subdir(share)
        is_local_share = self._smbutils.is_local_share(share)

        if not is_local_share:
            LOG.debug("Share '%s' is not exposed by the current host.", share)
            local_share_path = None
        else:
            local_share_path = self._smbutils.get_smb_share_path(share_name)

        if not local_share_path and expect_existing:
            err_msg = _("Could not find the local "
                        "share path for %(share)s.")
            raise exception.VolumePathsNotFound(err_msg % dict(share=share))

        if local_share_path and share_subdir:
            local_share_path = os.path.join(local_share_path, share_subdir)

        return local_share_path

    def _get_share_norm_path(self, share):
        return share.replace('/', '\\')

    def get_share_name(self, share):
        return self._get_share_norm_path(share).lstrip('\\').split('\\')[1]

    def get_share_subdir(self, share):
        return "\\".join(
            self._get_share_norm_path(share).lstrip('\\').split('\\')[2:])

    def mount(self, share, flags=None):
        share_norm_path = self._get_share_norm_path(share)
        use_local_path = (self._local_path_for_loopback and
                          self._smbutils.is_local_share(share_norm_path))

        if use_local_path:
            LOG.info("Skipping mounting local share %(share_path)s.",
                     dict(share_path=share_norm_path))
        else:
            mount_options = " ".join(
                [self._mount_options or '', flags or ''])
            username, password = self._parse_credentials(mount_options)

            if not self._smbutils.check_smb_mapping(
                    share_norm_path):
                self._smbutils.mount_smb_share(share_norm_path,
                                               username=username,
                                               password=password)

        if self._mount_base:
            self._create_mount_point(share, use_local_path)

    def unmount(self, share):
        self._smbutils.unmount_smb_share(self._get_share_norm_path(share))

    def _create_mount_point(self, share, use_local_path):
        # The mount point will contain a hash of the share so we're
        # intentionally preserving the original share path as this is
        # what the caller will expect.
        mnt_point = self.get_mount_point(share)
        share_norm_path = self._get_share_norm_path(share)
        symlink_dest = (share_norm_path if not use_local_path
                        else self.get_local_share_path(share))

        if not os.path.isdir(self._mount_base):
            os.makedirs(self._mount_base)

        if os.path.exists(mnt_point):
            if not self._pathutils.is_symlink(mnt_point):
                raise exception.BrickException(_("Link path already exists "
                                                 "and it's not a symlink"))
        else:
            self._pathutils.create_sym_link(mnt_point, symlink_dest)

    def _parse_credentials(self, opts_str):
        if not opts_str:
            return None, None

        match = self._username_regex.findall(opts_str)
        username = match[0] if match and match[0] != 'guest' else None

        match = self._password_regex.findall(opts_str)
        password = match[0] if match else None

        return username, password



# Copyright (c) 2013 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils

from os_brick.encryptors import nop

LOG = logging.getLogger(__name__)

LUKS = "luks"
LUKS2 = "luks2"
PLAIN = "plain"

FORMAT_TO_FRONTEND_ENCRYPTOR_MAP = {
    LUKS: 'os_brick.encryptors.luks.LuksEncryptor',
    LUKS2: 'os_brick.encryptors.luks.Luks2Encryptor',
    PLAIN: 'os_brick.encryptors.cryptsetup.CryptsetupEncryptor'
}

LEGACY_PROVIDER_CLASS_TO_FORMAT_MAP = {
    "nova.volume.encryptors.luks.LuksEncryptor": LUKS,
    "nova.volume.encryptors.cryptsetup.CryptsetupEncryptor": PLAIN,
    "nova.volume.encryptors.nop.NoopEncryptor": None,
    "os_brick.encryptors.luks.LuksEncryptor": LUKS,
    "os_brick.encryptors.cryptsetup.CryptsetupEncryptor": PLAIN,
    "os_brick.encryptors.nop.NoopEncryptor": None,
    "LuksEncryptor": LUKS,
    "CryptsetupEncryptor": PLAIN,
    "NoOpEncryptor": None,
}


def get_volume_encryptor(root_helper,
                         connection_info,
                         keymgr,
                         execute=None,
                         *args, **kwargs):
    """Creates a VolumeEncryptor used to encrypt the specified volume.

    :param: the connection information used to attach the volume
    :returns VolumeEncryptor: the VolumeEncryptor for the volume
    """
    encryptor = nop.NoOpEncryptor(root_helper=root_helper,
                                  connection_info=connection_info,
                                  keymgr=keymgr,
                                  execute=execute,
                                  *args, **kwargs)

    location = kwargs.get('control_location', None)
    if location and location.lower() == 'front-end':  # case insensitive
        provider = kwargs.get('provider')

        # TODO(lyarwood): Remove the following in Queens and raise an
        # ERROR if provider is not a key in SUPPORTED_ENCRYPTION_PROVIDERS.
        # Until then continue to allow both the class name and path to be used.
        if provider in LEGACY_PROVIDER_CLASS_TO_FORMAT_MAP:
            LOG.warning("Use of the in tree encryptor class %(provider)s"
                        " by directly referencing the implementation class"
                        " will be blocked in the Queens release of"
                        " os-brick.", {'provider': provider})
            provider = LEGACY_PROVIDER_CLASS_TO_FORMAT_MAP[provider]

        if provider in FORMAT_TO_FRONTEND_ENCRYPTOR_MAP:
            provider = FORMAT_TO_FRONTEND_ENCRYPTOR_MAP[provider]
        elif provider is None:
            provider = "os_brick.encryptors.nop.NoOpEncryptor"
        else:
            LOG.warning("Use of the out of tree encryptor class "
                        "%(provider)s will be blocked with the Queens "
                        "release of os-brick.", {'provider': provider})

        try:
            encryptor = importutils.import_object(
                provider,
                root_helper,
                connection_info,
                keymgr,
                execute,
                **kwargs)
        except Exception as e:
            LOG.error("Error instantiating %(provider)s: %(exception)s",
                      {'provider': provider, 'exception': e})
            raise

    msg = ("Using volume encryptor '%(encryptor)s' for connection: "
           "%(connection_info)s" %
           {'encryptor': encryptor, 'connection_info': connection_info})
    LOG.debug(strutils.mask_password(msg))

    return encryptor


def get_encryption_metadata(context, volume_api, volume_id, connection_info):
    metadata = {}
    if ('data' in connection_info and
            connection_info['data'].get('encrypted', False)):
        try:
            metadata = volume_api.get_volume_encryption_metadata(context,
                                                                 volume_id)
            if not metadata:
                LOG.warning('Volume %s should be encrypted but there is no '
                            'encryption metadata.', volume_id)
        except Exception as e:
            LOG.error("Failed to retrieve encryption metadata for "
                      "volume %(volume_id)s: %(exception)s",
                      {'volume_id': volume_id, 'exception': e})
            raise

    if metadata:
        msg = ("Using volume encryption metadata '%(metadata)s' for "
               "connection: %(connection_info)s" %
               {'metadata': metadata, 'connection_info': connection_info})
        LOG.debug(strutils.mask_password(msg))

    return metadata



# encoding=utf8
# (c) Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
from unittest import mock

from oslo_concurrency import processutils as putils
from oslo_context import context as context_utils

from os_brick import executor as brick_executor
from os_brick.privileged import rootwrap
from os_brick.tests import base


class TestExecutor(base.TestCase):
    def test_default_execute(self):
        executor = brick_executor.Executor(root_helper=None)
        self.assertEqual(rootwrap.execute, executor._Executor__execute)

    def test_none_execute(self):
        executor = brick_executor.Executor(root_helper=None, execute=None)
        self.assertEqual(rootwrap.execute, executor._Executor__execute)

    def test_fake_execute(self):
        mock_execute = mock.Mock()
        executor = brick_executor.Executor(root_helper=None,
                                           execute=mock_execute)
        self.assertEqual(mock_execute, executor._Executor__execute)

    @mock.patch('sys.stdin', encoding='UTF-8')
    @mock.patch('os_brick.executor.priv_rootwrap.execute')
    def test_execute_non_safe_str_exception(self, execute_mock, stdin_mock):
        execute_mock.side_effect = putils.ProcessExecutionError(
            stdout='España', stderr='Zürich')

        executor = brick_executor.Executor(root_helper=None)
        exc = self.assertRaises(putils.ProcessExecutionError,
                                executor._execute)
        self.assertEqual(u'Espa\xf1a', exc.stdout)
        self.assertEqual(u'Z\xfcrich', exc.stderr)

    @mock.patch('sys.stdin', encoding='UTF-8')
    @mock.patch('os_brick.executor.priv_rootwrap.execute')
    def test_execute_non_safe_str(self, execute_mock, stdin_mock):
        execute_mock.return_value = ('España', 'Zürich')

        executor = brick_executor.Executor(root_helper=None)
        stdout, stderr = executor._execute()
        self.assertEqual(u'Espa\xf1a', stdout)
        self.assertEqual(u'Z\xfcrich', stderr)

    @mock.patch('sys.stdin', encoding='UTF-8')
    @mock.patch('os_brick.executor.priv_rootwrap.execute')
    def test_execute_non_safe_bytes_exception(self, execute_mock, stdin_mock):
        execute_mock.side_effect = putils.ProcessExecutionError(
            stdout=bytes('España', 'utf-8'),
            stderr=bytes('Zürich', 'utf-8'))

        executor = brick_executor.Executor(root_helper=None)
        exc = self.assertRaises(putils.ProcessExecutionError,
                                executor._execute)
        self.assertEqual(u'Espa\xf1a', exc.stdout)
        self.assertEqual(u'Z\xfcrich', exc.stderr)

    @mock.patch('sys.stdin', encoding='UTF-8')
    @mock.patch('os_brick.executor.priv_rootwrap.execute')
    def test_execute_non_safe_bytes(self, execute_mock, stdin_mock):
        execute_mock.return_value = (bytes('España', 'utf-8'),
                                     bytes('Zürich', 'utf-8'))

        executor = brick_executor.Executor(root_helper=None)
        stdout, stderr = executor._execute()
        self.assertEqual(u'Espa\xf1a', stdout)
        self.assertEqual(u'Z\xfcrich', stderr)


class TestThread(base.TestCase):
    def _store_context(self, result):
        """Stores current thread's context in result list."""
        result.append(context_utils.get_current())

    def _run_threads(self, threads):
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    def _do_test(self, thread_class, expected, result=None):
        if result is None:
            result = []
        threads = [thread_class(target=self._store_context, args=[result])
                   for i in range(3)]
        self._run_threads(threads)
        self.assertEqual([expected] * len(threads), result)

    def test_normal_thread(self):
        """Test normal threads don't inherit parent's context."""
        context = context_utils.RequestContext()
        context.update_store()
        self._do_test(threading.Thread, None)

    def test_no_context(self, result=None):
        """Test when parent has no context."""
        context_utils._request_store.context = None
        self._do_test(brick_executor.Thread, None, result)

    def test_with_context(self, result=None):
        """Test that our class actually inherits the context."""
        context = context_utils.RequestContext()
        context.update_store()
        self._do_test(brick_executor.Thread, context, result)

    def _run_test(self, test_method, test_args, result):
        """Run one of the normal tests and store the result.

        Meant to be run in a different thread, thus the need to store the
        result, because by the time the join call completes the test's stack
        is no longer available and the exception will have been lost.
        """
        try:
            test_method(test_args)
            result.append(True)
        except Exception:
            result.append(False)
            raise

    def test_no_cross_mix(self):
        """Test there's no shared global context between threads."""
        result = []
        contexts = [[], [], []]
        threads = [threading.Thread(target=self._run_test,
                                    args=[self.test_with_context,
                                          contexts[0],
                                          result]),
                   threading.Thread(target=self._run_test,
                                    args=[self.test_no_context,
                                          contexts[1],
                                          result]),
                   threading.Thread(target=self._run_test,
                                    args=[self.test_with_context,
                                          contexts[2],
                                          result])]
        self._run_threads(threads)
        # Check that all tests run without raising an exception
        self.assertEqual([True, True, True], result)
        # Check that the context were not shared
        self.assertNotEqual(contexts[0], contexts[2])
